@@ -1,10 +1,11 @@
 /* [.vms]vms_popen.c -- substitute routines for missing pipe calls.
 
-   Copyright (C) 1991-1993, 1996, 2010, 2011 the Free Software Foundation, Inc.
+   Copyright (C) 1991-1993, 1996, 2010, 2011, 2014, 2022, 2023,
+   the Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
+   the Free Software Foundation; either version 3, or (at your option)
    any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -70,8 +71,8 @@ extern char *mktemp(char *);
 
 static void push_logicals(void);
 static void pop_logicals(void);
-static Itm *save_translation(const Dsc *);
-static void restore_translation(const Dsc *, const Itm *);
+static Itm *save_translation(const struct dsc$descriptor_s *);
+static void restore_translation(const struct dsc$descriptor_s *, const Itm *);
 
 typedef enum { unopened = 0, reading, writing } pipemode;
 typedef struct pipe_info {
@@ -133,8 +134,8 @@ pclose( FILE *current )
     int rval, cur = fileno(current);
 
  /* assert( cur >= 0 && cur < pipes_lim ); */
-    if (pipes[cur].pmode == unopened)
-	return -1;	/* should never happen */
+    if ((cur < 0) || (pipes[cur].pmode == unopened))
+	return -1;	/* should never happen, but does with two-way */
 
     rval = fclose(current);	/* close temp file; if reading, we're done */
     if (pipes[cur].pmode == writing) {
@@ -158,22 +159,30 @@ pclose( FILE *current )
 int
 vms_execute( const char *command, const char *input, const char *output )
 {
-    Dsc cmd, in, out, *in_p, *out_p;
+    struct dsc$descriptor_s cmd, in, out, *in_p, *out_p;
     U_Long sts, cmpltn_sts;
 
-    cmd.len = strlen(cmd.adr = (char *)command);
-    if (input)
-	in.len = strlen(in.adr = (char *)input),  in_p = &in;
-    else
+    cmd.dsc$w_length = strlen(cmd.dsc$a_pointer = (char *)command);
+    cmd.dsc$b_dtype = DSC$K_DTYPE_T;
+    cmd.dsc$b_class = DSC$K_CLASS_S;
+    if (input) {
+	in.dsc$w_length = strlen(in.dsc$a_pointer = (char *)input);
+	in_p = &in;
+	in.dsc$b_dtype = DSC$K_DTYPE_T;
+	in.dsc$b_class = DSC$K_CLASS_S;
+    } else
 	in_p = 0;
-    if (output)
-	out.len = strlen(out.adr = (char *)output),  out_p = &out;
-    else
+    if (output) {
+	out.dsc$w_length = strlen(out.dsc$a_pointer = (char *)output);
+	out_p = &out;
+	out.dsc$b_dtype = DSC$K_DTYPE_T;
+	out.dsc$b_class = DSC$K_CLASS_S;
+    } else
 	out_p = 0;
 
     push_logicals();	/* guard against user-mode definitions of sys$Xput */
-    sts = lib$spawn(&cmd, in_p, out_p, (U_Long *)0,
-		    (Dsc *)0, (U_Long *)0, &cmpltn_sts);
+    sts = LIB$SPAWN(&cmd, in_p, out_p, (U_Long *)0,
+		    (struct dsc$descriptor_s *)0, (U_Long *)0, &cmpltn_sts);
     pop_logicals();	/* restore environment */
 
     if (vmswork(sts) && vmsfail(cmpltn_sts))  sts = cmpltn_sts;
@@ -219,9 +228,9 @@ static const Descrip(sys_output,"SYS$OUTPUT");
 static const unsigned char acmode = PSL$C_USER; /* only care about user-mode */
 
  /* macros for simplfying the code a bunch */
-#define DelTrans(l)	sys$dellnm(&lnmtable, (l), &acmode)
-#define GetTrans(l,i)	sys$trnlnm((U_Long *)0, &lnmtable, (l), &acmode, (i))
-#define SetTrans(l,i)	sys$crelnm((U_Long *)0, &lnmtable, (l), &acmode, (i))
+#define DelTrans(l)	SYS$DELLNM(&lnmtable, (l), &acmode)
+#define GetTrans(l,i)	SYS$TRNLNM((U_Long *)0, &lnmtable, (l), &acmode, (i))
+#define SetTrans(l,i)	SYS$CRELNM((U_Long *)0, &lnmtable, (l), &acmode, (i))
  /* itemlist manipulation macros; separate versions for aggregate and scalar */
 #define SetItmA(i,c,p,r) ((i).code = (c), (i).len = sizeof (p),\
 			  (i).buffer = (p), (i).retlen = (U_Short *)(r))
@@ -254,7 +263,7 @@ pop_logicals( void )		/* redefine sys$input and/or sys$output */
 }
 
 static Itm *
-save_translation( const Dsc *logname )
+save_translation( const struct dsc$descriptor_s *logname )
 {
     Itm trans[4], *itmlst;
     long trans_attr, max_trans_indx;	/* 0-based translation index count */
@@ -317,9 +326,10 @@ save_translation( const Dsc *logname )
 }
 
 static void
-restore_translation( const Dsc *logname, const Itm *itemlist )
+restore_translation( const struct dsc$descriptor_s *logname,
+                     const Itm *itemlist )
 {
-    Dsc trans_val;
+    struct dsc$descriptor_s trans_val;
     U_Long *attr_p;
 # define LOG_PROCESS_TABLE 2		/* <obsolete> */
 # define LOG_USERMODE PSL$C_USER
@@ -331,9 +341,11 @@ restore_translation( const Dsc *logname, const Itm *itemlist )
 	    so it'll be the first string entry in the itemlist.
 	 */
      /* assert( itemlist[2].code == LNM$_STRING ); */
-	trans_val.adr = itemlist[2].buffer;
-	trans_val.len = itemlist[2].len;
-	(void) sys$crelog(LOG_PROCESS_TABLE, logname, &trans_val, LOG_USERMODE);
+	trans_val.dsc$a_pointer = itemlist[2].buffer;
+	trans_val.dsc$w_length = itemlist[2].len;
+	trans_val.dsc$b_dtype = DSC$K_DTYPE_T;
+	trans_val.dsc$b_class = DSC$K_CLASS_S;
+	(void) SYS$CRELOG(LOG_PROCESS_TABLE, logname, &trans_val, LOG_USERMODE);
     } else {
 	/* $crelnm definition; itemlist could specify multiple translations,
 	    but has already been setup properly for use as-is.
